// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package xbtree complements package lab.nexedi.com/kirr/neo/go/zodb/btree.
//
// It provides the following amendments:
//
// - ΔBtail (tail of revisional changes to BTrees).
package xbtree


import (
	"fmt"

	"lab.nexedi.com/kirr/neo/go/zodb"

	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/set"
	"lab.nexedi.com/nexedi/wendelin.core/wcfs/internal/xbtree/blib"
)


// XXX instead of generics
type Tree   = blib.Tree
type Bucket = blib.Bucket
type Node   = blib.Node
type TreeEntry   = blib.TreeEntry
type BucketEntry = blib.BucketEntry

type Key    = blib.Key
const KeyMax = blib.KeyMax
const KeyMin = blib.KeyMin

// value is assumed to be persistent reference.
// deletion is represented as VDEL.
type Value  = zodb.Oid
const VDEL  = zodb.InvalidOid

type setKey = set.I64
type setOid = set.Oid
type setTid = set.Tid




// nodePathToPath converts path from []Node to []Oid.
// XXX place = ?  -> δbtail.go/treediff.go (close to pathEqual), or move them both to misc.go
func nodePathToPath(nodePath []Node) (path []zodb.Oid) {
	// assert nodePath = Tree Tree ... Tree Bucket
	l := len(nodePath)
	switch {
	case l == 0:
		panic("empty path")
	case l == 1:
		// must be empty Tree
		_ = nodePath[0].(*Tree)
	default:
		// must be Tree Tree ... Tree Bucket
		for _, node := range nodePath[:l-1] {
			_ = node.(*Tree)
		}
		_ = nodePath[l-1].(*Bucket)
	}

	path = make([]zodb.Oid, l)
	for i, node := range nodePath {
		path[i] = node.POid()
	}
	return path
}


// kstr formats key as string.
func kstr(k Key) string {
	if k == KeyMin {
		return "-∞"
	}
	if k == KeyMax {
		return "∞"
	}
	return fmt.Sprintf("%d", k)
}

func panicf(format string, argv ...interface{}) {
	panic(fmt.Sprintf(format, argv...))
}
